/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>
#include <list>
#include <string>
#include <sstream>
#include <fstream>

#include "defin.h"
#include "commandline.h"
#include "executable.h"
#include "client.h"
#include "receivor.h"
#include "runprocess.h"
#include "modelrun.h"
#include "modelresult.h"
#include "linpack_bench.h"
#include "restart.h"

using namespace std;

//****************************************************************************
int main(int argc,     //number of args in argv
         char *argv[]) // Array of arguments
//****************************************************************************

  // C.Muffels, S.S. Papadopulos & Assoc. Inc.
  // W.Shreuder, Principia Mathematica
  // Inspired by a run manager by J.Doherty and BeoPEST by W.Shreuder

  // Slave (or Run Executor): Communication from HOST is managed by a RECEIVOR
  //                          thread. The primary thread is responsible for
  //                          managing run processes and sending messages to
  //                          the HOST.

  // Version 1.0.0.0  Jan. 2011

{

  string lowercase(string s);

  //--- Restart parameters
  RESTART restart_todo;
  RESTART restart_inprogress;
  RESTART restart_tosend;
  RESTART restart_pid;

  //--- Misc
//  int stringtoint(string &_string);
  //bool endPROGRAM=false;
  HANDLE myhandle;
  COMMANDLINE *cmd;
  string *stmp;
  double interval=1.;
  bool *console_visible;

  //--- Speed check
  LINPACK_BENCH lpb;

  //--- Event triggers
  unsigned int event_wait_result;
  unsigned int mutex_wait_result;

  //--- Model

  //--- Slave
  CLIENT slave;

  MODEL_RESULT *result;

  //--- Thread
  RECEIVOR *thread;
  HANDLE h_thEnd=NULL;
  HANDLE h_thComm=NULL;
  HANDLE m_thRuns=NULL;
  //unsigned long threadstatus=0;

  //--- Run
  int runs_complete=0;
  list<MODEL_RUN*> *runs_new;
  list<MODEL_RUN*> *runs_todo;
  list<MODEL_RUN*> *runs_inprogress;
  list<MODEL_RUN*>::iterator run;
  RUNPROCESS *runner;
  unsigned long processstatus;

  //--- Socket
  WSADATA wsaData;

  // Initialize pointers
  runs_new=new list<MODEL_RUN*>;
  runs_todo=new list<MODEL_RUN*>;
  runs_inprogress=new list<MODEL_RUN*>;
  thread=new RECEIVOR;
  runner=new RUNPROCESS;

  // Create events/mutexes
  SECURITY_DESCRIPTOR sd={0};
  SECURITY_ATTRIBUTES saAttr;

  InitializeSecurityDescriptor(&sd,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&sd,true,NULL,false);
  saAttr.nLength=sizeof(SECURITY_ATTRIBUTES);
  saAttr.lpSecurityDescriptor=&sd;
  saAttr.bInheritHandle=true;

  h_thComm=CreateEvent(&saAttr,false,false,"ThreadComm");
  if(h_thComm==NULL) throw ERROR_EVENT_INIT;
  h_thEnd=CreateEvent(&saAttr,true,false,"ThreadEnd");
  if(h_thEnd==NULL) throw ERROR_EVENT_INIT;
  m_thRuns=CreateMutex(&saAttr,false,"RunsQueue");
  if(m_thRuns==NULL) throw ERROR_EVENT_INIT;

  myhandle=GetCurrentProcess();

  // get main routine specific command line args
  cmd=new COMMANDLINE;
  stmp=new string;
  console_visible=new bool;
  *console_visible=false;
  //pass commandline arguments to cmd
  cmd->setarg(argc,argv);
  //get the "wait" interval
  *stmp=cmd->getarg("/interval");
  if(stmp->compare("NOT_FOUND")==0)
    interval=COMM_WAIT_TIME;
  else
    interval=atof(stmp->c_str());
  //check the console flag
  *stmp=cmd->getarg("/console");
  cout<<*stmp<<'\n';
  if(lowercase(*stmp).compare("on")==0)
    *console_visible=true;
  //delete variables
  delete cmd;
  delete stmp;

  try
  {

    // initialize winsock
    if(WSAStartup(WINSOCK_VERSION,&wsaData)!=0) throw ERROR_WSASTARTUP;

    // initialize and start slave
    if(!slave.initialize(argc,argv)) throw ERROR_NODE_INIT;
    if(!slave.start()) throw ERROR_NODE_START;
    // open restart files
    //cout<<"Opening RESTART files..."<<'\n';
    //cout<<"   TODO.GRF... ";
    //if(!restart_todo.open_me("todo.grf")) throw ERROR_RESTART_OPEN;
    //cout<<"OK"<<'\n';
    //cout<<"   INPROGRESS.GRF... ";
    //if(!restart_inprogress.open_me("inprogress.grf")) throw ERROR_RESTART_OPEN;
    //cout<<"OK"<<'\n';
    //cout<<"   TOSEND.GRF... ";
    //if(!restart_tosend.open_me("tosend.grf")) throw ERROR_RESTART_OPEN;
    //cout<<"OK"<<'\n';
    //cout<<"   PID.GRF... ";
    //if(!restart_pid.open_me("pid.grf")) throw ERROR_RESTART_OPEN;
    //cout<<"OK"<<'\n';

    // check if there are any runs inprogress...
    // if there is a run in progress - get the PID and see if the process is still running
    // if the process is still running set status and go into gslave as normal
    // if the process is no longer running - check if the output files associated with this
    //   run exist - if they do then the run is finished - get the results and return to GMAN
    // if the output files do not exist then the run did not finish and the model terminated
    //   unexpectedly, so, restart this run and norify GMAN as such
    // if there are no runs in progress - check if there is a run to send... if so, send it and move on
    // if there are no runs to send then read the runs todo and proceed as normal
    // if there are no runs to do then start gslave as normal

    // send name to host
    cout<<"\n   Sending NAME to GMAN... ";
    if(!slave.sendname()) throw ERROR_NODE_SENDNAME;
    cout<<"Complete."<<'\n';

    // get baseline speed using Linpack Benchmark and send to host
    slave.rating=lpb.mflops();
    if(slave.rating==-999.) throw ERROR_GET_SPEED;
    cout<<"   Speed rating: "<<slave.rating<<'\n';
    cout<<"   Sending RATING to GMAN... ";
    if(!slave.sendspeed()) throw ERROR_NODE_SENDSPEED;
    cout<<"Complete."<<'\n';

    // send slave type to host
    cout<<"   Sending TYPE to GMAN... ";
    if(!slave.sendtype()) throw ERROR_NODE_SENDTYPE;
    cout<<"Complete."<<'\n';

    // associate slave with thread and start.
    // this thread will handle communication with the host.
    cout<<"   Initialize THREAD to handle communication with GMAN... ";
    if(!thread->initialize()) throw ERROR_THREAD_START;
    thread->set_primary(myhandle);
    thread->set_slave(&slave);
    thread->set_queue_runs(runs_new);
    thread->set_event_comm(h_thComm);
    thread->set_mutex_comm(m_thRuns);
    thread->start();
    cout<<"Complete."<<'\n';

    // initialize run process
    cout<<"   Initialize RUN PROCESS ... ";
    runner->initialize(console_visible);
    cout<<"Complete."<<'\n';

    // wait for something to happen!
    while(1)
    {
      //cout<<" -----------------------------------------------------------"<<'\n';
      //cout<<"   No. of Runs Remaining: "<<runs_todo->size()<<'\n';
      //cout<<"   No. of Runs In-progress: "<<runs_inprogress->size()<<'\n';
      //cout<<"   No. of Runs Complete: "<<runs_complete<<'\n';
      //cout<<" -----------------------------------------------------------"<<'\n';
      event_wait_result=WaitForSingleObject(h_thComm,interval*CLOCKS_PER_SEC);
      switch(event_wait_result)
      {
      case WAIT_FAILED:
        throw ERROR_EVENT_WAIT;
        break;
      default:
        // check if slave/run is to be killed
        if(slave.status==KILL_RUN||slave.status==KILL_ALL) //||slave->status==KILL_SLAVE)
        {
          //ensure process is active
          if(!GetExitCodeProcess(runner->pinfo.hProcess,&processstatus)) break;
          switch(processstatus)
          {
          case STILL_ACTIVE:
            cout<<"\n   Killing RUN."<<'\n';
            if(!runner->KillProcessTree(runner->pinfo.dwProcessId)) throw ERROR_RUN_KILL;
            CloseHandle(runner->pinfo.hProcess);
            CloseHandle(runner->pinfo.hThread);
            runs_inprogress->pop_front();
            // remove run from restart file
            break;
          }
        }
        // check if slave is to be terminated
        if(slave.status==KILL_SLAVE||slave.status==KILL_ALL)
          throw NORMAL_TERMINATION;
  
        // check if there are new runs
        mutex_wait_result=WaitForSingleObject(m_thRuns,0);
        if(mutex_wait_result==WAIT_OBJECT_0)
        {
          run=runs_new->begin();
          while(run!=runs_new->end())
          {
            cout<<"\n   Have new RUN: "<<*(*run)->id<<'\n';
            restart_todo.put_run(*run);
            runs_todo->push_back(*run);
            run=runs_new->erase(run);
          }
          if(!ReleaseMutex(m_thRuns)) throw ERROR_EVENT_RELEASE;
        }

        // check if any runs are complete
        run=runs_inprogress->begin();
        while(run!=runs_inprogress->end())
        {
          //cout<<"\n\n Checking whether process is still active."<<'\n';
          if(!GetExitCodeProcess(runner->pinfo.hProcess,&processstatus)) break;
          switch(processstatus)
          {
          case STILL_ACTIVE:
            ++run;
            break;
          case PROCESS_COMPLETE:
            // process run and request another
            cout<<"\n   Completed RUN: "<<*(*run)->id<<'\n';
            CloseHandle(runner->pinfo.hProcess);
            CloseHandle(runner->pinfo.hThread);
            // get results
            cout<<"   - Processing results... ";
            if(!(*run)->read_output()) throw ERROR_RUN_READ_OUT;
            cout<<"Complete."<<'\n';
            // send results to manager
            cout<<"   - Sending results to GMAN... ";
            result=new MODEL_RESULT;
            result->set_frm_run(*run);
            restart_tosend.put_run(*run);
            if(!result->send(slave.sock)) throw ERROR_NODE_SENDOBS;
            cout<<"Complete."<<'\n';
            (*run)->dealloc();
            delete *run;
            delete result;
            runs_complete++;
            run=runs_inprogress->erase(run);
            break;
          default:
            cout<<"   Error executing RUN: "<<(*run)->id<<'\n';
            // error executing run
            CloseHandle(runner->pinfo.hProcess);
            CloseHandle(runner->pinfo.hThread);
            ++run;
            break;
          }
        }

        // check if there are any runs to do
        run=runs_todo->begin();
        if(run!=runs_todo->end())
        {
          // update run input files
          cout<<"   - Writing RUN inputs... ";
          if(!(*run)->write_input()) throw ERROR_RUN_WRITE_IN;
          cout<<"Complete."<<'\n';
          (*run)->delete_output();
          runner->set_executable((*run)->exec);
          //cout<<"Starting model run."<<'\n';
          if(!runner->start()) throw ERROR_RUN_START;
          // add run to restart file
          //if(!restart_inprogress.put_run(*run)) throw ERROR_RESTART_WRITE;
          runs_inprogress->push_back(*run);
          run=runs_todo->erase(run);
          // remove run from todo restart file
        }

        break;
      }
    }

    restart_inprogress.close_me();
    throw NORMAL_TERMINATION;
  }

  catch (int err)
  {

    cout<<"\n\n *** Terminating Program... **********\n"<<'\n';

    // Kill the RECEIVOR thread
    if(*thread->iset)
    {
      cout<<" *** Terminating RECEIVOR thread."<<'\n';
      if(SetEvent(h_thEnd))
      {
        cout<<"     Waiting for thread to terminate."<<'\n';
        WaitForSingleObject(thread->handle,INFINITE);
        CloseHandle(thread->handle);
        cout<<"     Thread terminated normally."<<'\n';
        slave.stop();
        cout<<"     Connection with HOST is closed."<<'\n';
      }
      else
      {
        // force termination of thread
        TerminateThread(thread->handle,1);
        cout<<"     Forcing termination of thread."<<'\n';
        slave.stop();
        CloseHandle(thread->handle);
        CloseHandle(h_thEnd);
        cout<<"     Connection with HOST is closed.";
      }
    }
    else
      CloseHandle(h_thEnd);
    delete thread;

    // Terminate run process
    if(*runner->iset)
    {
      cout<<" *** Terminating RUN PROCESS."<<'\n';
      // check if process is active
      GetExitCodeProcess(runner->pinfo.hProcess,&processstatus);
      if(processstatus==STILL_ACTIVE)
        runner->KillProcessTree(runner->pinfo.dwProcessId);
      CloseHandle(runner->pinfo.hProcess);
      CloseHandle(runner->pinfo.hThread);
    }
    delete runner;

    // Close event handles
    cout<<" *** Closing EVENT handles."<<'\n';
    CloseHandle(h_thComm);
    CloseHandle(m_thRuns);

    // Clear run queues
    cout<<" *** Clearing RUN queues."<<'\n';
    runs_new->clear();
    delete runs_new;
    runs_todo->clear();
    delete runs_todo;
    runs_inprogress->clear();
    delete runs_inprogress;

    if(WSACleanup()==SOCKET_ERROR) err=ERROR_WSACLEANUP;
     switch(err)
    {
    case NORMAL_TERMINATION:
      cout<<"\n\n Normal Termination of Program."<<'\n';
      return 1;
      break;
    case ERROR_WSACLEANUP:
      cout<<"\n\n Winsock function 'WSACleanup' failed with error: "<<WSAGetLastError()<<'\n';
      break;
    case ERROR_WSASTARTUP:
      cout<<"\n\n Winsock function 'WSAStartup' failed with error: "<<WSAGetLastError()<<'\n';
      break;
    case ERROR_NODE_INIT:
      cout<<"\n *** Error initializing this NODE."<<'\n';
      break;
    case ERROR_NODE_START:
      cout<<"\n *** Error starting this NODE."<<'\n';
      break;
    case ERROR_NODE_SENDNAME:
      cout<<"\n *** Error sending name of this NODE to HOST."<<'\n';
      break;
    case ERROR_NODE_SENDTYPE:
      cout<<"\n *** Error sending type of this NODE to HOST."<<'\n';
      break;
    case ERROR_NODE_SENDSPEED:
      cout<<"\n *** Error sending MFLOPS rating of this NODE to HOST."<<'\n';
      break;
    case ERROR_NODE_SENDSTATUS:
      cout<<"\n *** Error sending status of this NODE to HOST."<<'\n';
      break;
    case ERROR_THREAD_START:
      cout<<"\n *** Error starting RECEIVOR thread."<<'\n';
      break;
    case ERROR_EVENT_INIT:
      cout<<"\n *** Error initializing EVENT"<<'\n';
      cout<<"     (communication from host)."<<'\n';
      break;
    case ERROR_EVENT_WAIT:
      cout<<"\n *** Unexpected error waiting for EVENT"<<'\n';
      cout<<"     (communication from host)."<<'\n';
      break;
    case ERROR_THREAD_CHKEXIT:
      cout<<"\n\n Error getting exit code of thread. No suggestions except to restart."<<'\n';
      break;
    case ERROR_THREAD_EXEC:
      cout<<"\n *** Error with thread execution."<<'\n';
      break;
    case ERROR_GET_SPEED:
      cout<<"\n *** Error determining MFLOPS rating of this computer."<<'\n';
      break;
    case ERROR_RUN_KILL:
      cout<<"\n *** Error terminating run."<<'\n';
      break;
    case ERROR_RESTART_OPEN:
      cout<<"\n *** Error opening restart file."<<'\n';
      break;
    case ERROR_RESTART_WRITE:
      cout<<"\n *** Error writing to restart file."<<'\n';
      break;
    }
    cout<<"\n *** Termination of program with errors."<<'\n';

   return 0;
  }

}

//*****************************************************************************
string lowercase(string s)
//*****************************************************************************
//CTM, Nov 2010

/*
Routine to convert string to lower case
*/

{
  for(unsigned int i=0;i<s.size();i++) {
    s[i]=tolower(s[i]);
  }
  return s;
}